# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model-runtime.proto
"""Generated protocol buffer code."""
# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13model-runtime.proto\x12\x05mmesh"[\n\x10LoadModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12\x11\n\tmodelType\x18\x02 \x01(\t\x12\x11\n\tmodelPath\x18\x03 \x01(\t\x12\x10\n\x08modelKey\x18\x04 \x01(\t"@\n\x11LoadModelResponse\x12\x13\n\x0bsizeInBytes\x18\x01 \x01(\x04\x12\x16\n\x0emaxConcurrency\x18\x02 \x01(\r"%\n\x12UnloadModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"\x15\n\x13UnloadModelResponse"b\n\x17PredictModelSizeRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12\x11\n\tmodelType\x18\x02 \x01(\t\x12\x11\n\tmodelPath\x18\x03 \x01(\t\x12\x10\n\x08modelKey\x18\x04 \x01(\t"/\n\x18PredictModelSizeResponse\x12\x13\n\x0bsizeInBytes\x18\x01 \x01(\x04"#\n\x10ModelSizeRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"(\n\x11ModelSizeResponse\x12\x13\n\x0bsizeInBytes\x18\x01 \x01(\x04"\x16\n\x14RuntimeStatusRequest"\xaa\x04\n\x15RuntimeStatusResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.mmesh.RuntimeStatusResponse.Status\x12\x17\n\x0f\x63\x61pacityInBytes\x18\x02 \x01(\x04\x12\x1d\n\x15maxLoadingConcurrency\x18\x03 \x01(\r\x12\x1d\n\x15modelLoadingTimeoutMs\x18\x04 \x01(\r\x12\x1f\n\x17\x64\x65\x66\x61ultModelSizeInBytes\x18\x05 \x01(\x04\x12\x16\n\x0eruntimeVersion\x18\x06 \x01(\t\x12\x1d\n\x15numericRuntimeVersion\x18\x07 \x01(\x04\x12\x42\n\x0bmethodInfos\x18\x08 \x03(\x0b\x32-.mmesh.RuntimeStatusResponse.MethodInfosEntry\x12\x1d\n\x15limitModelConcurrency\x18\t \x01(\x08\x12\x16\n\x0e\x61llowAnyMethod\x18\n \x01(\x08\x1a%\n\nMethodInfo\x12\x17\n\x0fidInjectionPath\x18\x01 \x03(\r\x1a[\n\x10MethodInfosEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.mmesh.RuntimeStatusResponse.MethodInfo:\x02\x38\x01".\n\x06Status\x12\x0c\n\x08STARTING\x10\x00\x12\t\n\x05READY\x10\x01\x12\x0b\n\x07\x46\x41ILING\x10\x02\x32\xff\x02\n\x0cModelRuntime\x12@\n\tloadModel\x12\x17.mmesh.LoadModelRequest\x1a\x18.mmesh.LoadModelResponse"\x00\x12\x46\n\x0bunloadModel\x12\x19.mmesh.UnloadModelRequest\x1a\x1a.mmesh.UnloadModelResponse"\x00\x12U\n\x10predictModelSize\x12\x1e.mmesh.PredictModelSizeRequest\x1a\x1f.mmesh.PredictModelSizeResponse"\x00\x12@\n\tmodelSize\x12\x17.mmesh.ModelSizeRequest\x1a\x18.mmesh.ModelSizeResponse"\x00\x12L\n\rruntimeStatus\x12\x1b.mmesh.RuntimeStatusRequest\x1a\x1c.mmesh.RuntimeStatusResponse"\x00\x42 \n\x1c\x63om.ibm.watson.modelmesh.apiP\x01\x62\x06proto3'
)


_LOADMODELREQUEST = DESCRIPTOR.message_types_by_name["LoadModelRequest"]
_LOADMODELRESPONSE = DESCRIPTOR.message_types_by_name["LoadModelResponse"]
_UNLOADMODELREQUEST = DESCRIPTOR.message_types_by_name["UnloadModelRequest"]
_UNLOADMODELRESPONSE = DESCRIPTOR.message_types_by_name["UnloadModelResponse"]
_PREDICTMODELSIZEREQUEST = DESCRIPTOR.message_types_by_name["PredictModelSizeRequest"]
_PREDICTMODELSIZERESPONSE = DESCRIPTOR.message_types_by_name["PredictModelSizeResponse"]
_MODELSIZEREQUEST = DESCRIPTOR.message_types_by_name["ModelSizeRequest"]
_MODELSIZERESPONSE = DESCRIPTOR.message_types_by_name["ModelSizeResponse"]
_RUNTIMESTATUSREQUEST = DESCRIPTOR.message_types_by_name["RuntimeStatusRequest"]
_RUNTIMESTATUSRESPONSE = DESCRIPTOR.message_types_by_name["RuntimeStatusResponse"]
_RUNTIMESTATUSRESPONSE_METHODINFO = _RUNTIMESTATUSRESPONSE.nested_types_by_name[
    "MethodInfo"
]
_RUNTIMESTATUSRESPONSE_METHODINFOSENTRY = _RUNTIMESTATUSRESPONSE.nested_types_by_name[
    "MethodInfosEntry"
]
_RUNTIMESTATUSRESPONSE_STATUS = _RUNTIMESTATUSRESPONSE.enum_types_by_name["Status"]
LoadModelRequest = _reflection.GeneratedProtocolMessageType(
    "LoadModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LOADMODELREQUEST,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.LoadModelRequest)
    },
)
_sym_db.RegisterMessage(LoadModelRequest)

LoadModelResponse = _reflection.GeneratedProtocolMessageType(
    "LoadModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _LOADMODELRESPONSE,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.LoadModelResponse)
    },
)
_sym_db.RegisterMessage(LoadModelResponse)

UnloadModelRequest = _reflection.GeneratedProtocolMessageType(
    "UnloadModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNLOADMODELREQUEST,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnloadModelRequest)
    },
)
_sym_db.RegisterMessage(UnloadModelRequest)

UnloadModelResponse = _reflection.GeneratedProtocolMessageType(
    "UnloadModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNLOADMODELRESPONSE,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnloadModelResponse)
    },
)
_sym_db.RegisterMessage(UnloadModelResponse)

PredictModelSizeRequest = _reflection.GeneratedProtocolMessageType(
    "PredictModelSizeRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PREDICTMODELSIZEREQUEST,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.PredictModelSizeRequest)
    },
)
_sym_db.RegisterMessage(PredictModelSizeRequest)

PredictModelSizeResponse = _reflection.GeneratedProtocolMessageType(
    "PredictModelSizeResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _PREDICTMODELSIZERESPONSE,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.PredictModelSizeResponse)
    },
)
_sym_db.RegisterMessage(PredictModelSizeResponse)

ModelSizeRequest = _reflection.GeneratedProtocolMessageType(
    "ModelSizeRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELSIZEREQUEST,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelSizeRequest)
    },
)
_sym_db.RegisterMessage(ModelSizeRequest)

ModelSizeResponse = _reflection.GeneratedProtocolMessageType(
    "ModelSizeResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELSIZERESPONSE,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelSizeResponse)
    },
)
_sym_db.RegisterMessage(ModelSizeResponse)

RuntimeStatusRequest = _reflection.GeneratedProtocolMessageType(
    "RuntimeStatusRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _RUNTIMESTATUSREQUEST,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.RuntimeStatusRequest)
    },
)
_sym_db.RegisterMessage(RuntimeStatusRequest)

RuntimeStatusResponse = _reflection.GeneratedProtocolMessageType(
    "RuntimeStatusResponse",
    (_message.Message,),
    {
        "MethodInfo": _reflection.GeneratedProtocolMessageType(
            "MethodInfo",
            (_message.Message,),
            {
                "DESCRIPTOR": _RUNTIMESTATUSRESPONSE_METHODINFO,
                "__module__": "model_runtime_pb2"
                # @@protoc_insertion_point(class_scope:mmesh.RuntimeStatusResponse.MethodInfo)
            },
        ),
        "MethodInfosEntry": _reflection.GeneratedProtocolMessageType(
            "MethodInfosEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _RUNTIMESTATUSRESPONSE_METHODINFOSENTRY,
                "__module__": "model_runtime_pb2"
                # @@protoc_insertion_point(class_scope:mmesh.RuntimeStatusResponse.MethodInfosEntry)
            },
        ),
        "DESCRIPTOR": _RUNTIMESTATUSRESPONSE,
        "__module__": "model_runtime_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.RuntimeStatusResponse)
    },
)
_sym_db.RegisterMessage(RuntimeStatusResponse)
_sym_db.RegisterMessage(RuntimeStatusResponse.MethodInfo)
_sym_db.RegisterMessage(RuntimeStatusResponse.MethodInfosEntry)

_MODELRUNTIME = DESCRIPTOR.services_by_name["ModelRuntime"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\034com.ibm.watson.modelmesh.apiP\001"
    _RUNTIMESTATUSRESPONSE_METHODINFOSENTRY._options = None
    _RUNTIMESTATUSRESPONSE_METHODINFOSENTRY._serialized_options = b"8\001"
    _LOADMODELREQUEST._serialized_start = 30
    _LOADMODELREQUEST._serialized_end = 121
    _LOADMODELRESPONSE._serialized_start = 123
    _LOADMODELRESPONSE._serialized_end = 187
    _UNLOADMODELREQUEST._serialized_start = 189
    _UNLOADMODELREQUEST._serialized_end = 226
    _UNLOADMODELRESPONSE._serialized_start = 228
    _UNLOADMODELRESPONSE._serialized_end = 249
    _PREDICTMODELSIZEREQUEST._serialized_start = 251
    _PREDICTMODELSIZEREQUEST._serialized_end = 349
    _PREDICTMODELSIZERESPONSE._serialized_start = 351
    _PREDICTMODELSIZERESPONSE._serialized_end = 398
    _MODELSIZEREQUEST._serialized_start = 400
    _MODELSIZEREQUEST._serialized_end = 435
    _MODELSIZERESPONSE._serialized_start = 437
    _MODELSIZERESPONSE._serialized_end = 477
    _RUNTIMESTATUSREQUEST._serialized_start = 479
    _RUNTIMESTATUSREQUEST._serialized_end = 501
    _RUNTIMESTATUSRESPONSE._serialized_start = 504
    _RUNTIMESTATUSRESPONSE._serialized_end = 1058
    _RUNTIMESTATUSRESPONSE_METHODINFO._serialized_start = 880
    _RUNTIMESTATUSRESPONSE_METHODINFO._serialized_end = 917
    _RUNTIMESTATUSRESPONSE_METHODINFOSENTRY._serialized_start = 919
    _RUNTIMESTATUSRESPONSE_METHODINFOSENTRY._serialized_end = 1010
    _RUNTIMESTATUSRESPONSE_STATUS._serialized_start = 1012
    _RUNTIMESTATUSRESPONSE_STATUS._serialized_end = 1058
    _MODELRUNTIME._serialized_start = 1061
    _MODELRUNTIME._serialized_end = 1444
# @@protoc_insertion_point(module_scope)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""These are mostly some older tests that we wrote ad-hoc while throwing the inference proto-generation together

Coverage is probably not the best
"""
# Standard
from typing import List, Optional
from unittest.mock import patch
import inspect

# Third Party
from docstring_parser import ParseError

# Local
from caikit.core.signature_parsing.parsers import (
    _get_dm_type_from_name,
    _snake_to_camel,
    get_args_with_defaults,
    get_argument_types,
    get_output_type_name,
)
import caikit.core
import sample_lib

## Tests ########################################################################

# TODO: Add a test that looks for a type in caikit.interfaces.common.data_model


def test_snake_to_camel():
    assert _snake_to_camel("this_is_a_test_str") == "ThisIsATestStr"


def test_get_dm_type_from_name():
    assert _get_dm_type_from_name(None) == None

    assert (
        _get_dm_type_from_name("ProducerId")
        == caikit.core.data_model.producer.ProducerId
    )

    assert (
        _get_dm_type_from_name("SampleOutputType")
        == sample_lib.data_model.sample.SampleOutputType
    )

    assert _get_dm_type_from_name("NonExistentName") == None


def test_get_output_type_name():

    # Test that if there's no function signature, use docstring to deduct output type
    empty_sign = inspect.Signature(return_annotation=inspect.Signature.empty)
    assert (
        get_output_type_name(
            module_class=sample_lib.modules.sample_task.SampleModule,
            fn_signature=empty_sign,
            fn=sample_lib.modules.sample_task.SampleModule.run,
        )
        == sample_lib.data_model.SampleOutputType
    )

    # Test that we use type annotation to deduct output type
    inner_module_run_method_ptr = getattr(
        sample_lib.modules.sample_task.InnerModule, "run"
    )
    fn_sign = inspect.signature(inner_module_run_method_ptr)
    assert (
        get_output_type_name(
            module_class=sample_lib.modules.sample_task.InnerModule,
            fn_signature=fn_sign,
            fn=sample_lib.modules.sample_task.InnerModule.run,
        )
        == sample_lib.data_model.SampleOutputType
    )

    # Test that we use type annotation to deduct output type is return annotation is a string
    def _run(self, some_input: str) -> "InnerModule":
        pass

    fn_sign = inspect.signature(_run)
    assert (
        get_output_type_name(
            module_class=sample_lib.modules.sample_task.InnerModule,
            fn_signature=fn_sign,
            fn=sample_lib.modules.sample_task.InnerModule.run,
        )
        == sample_lib.modules.sample_task.InnerModule
    )

    # Test that we return None if type annotation as a string that doesn't match module class name
    def _run2(self, some_input: str) -> "AStringThatsNotInnerModule":
        pass

    fn_sign = inspect.signature(_run2)
    assert (
        get_output_type_name(
            module_class=sample_lib.modules.sample_task.InnerModule,
            fn_signature=fn_sign,
            fn=sample_lib.modules.sample_task.InnerModule.run,
        )
        == None
    )

    # User doesn't provide any type annotation or docstring, return None
    assert (
        get_output_type_name(
            module_class=sample_lib.modules.sample_task.InnerModule,
            fn_signature=empty_sign,
            fn=sample_lib.modules.sample_task.InnerModule.run,
        )
        == None
    )

    # Test that if a ParseError was raised with docstring.parsers, we have no idea the type and return None
    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert (
            get_output_type_name(
                module_class=sample_lib.modules.sample_task.InnerModule,
                fn_signature=empty_sign,
                fn=sample_lib.modules.sample_task.InnerModule.run,
            )
            == None
        )


def test_get_argument_types_with_real_module():
    """Quick check that we get the right type for our sample module"""
    assert (
        get_argument_types(sample_lib.modules.sample_task.SampleModule.run)[
            "sample_input"
        ]
        == sample_lib.data_model.SampleInputType
    )

    # Test that if a ParseError was raised with docstring.parsers, we could still parse from type annotation
    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert (
            get_argument_types(sample_lib.modules.sample_task.SampleModule.run)[
                "sample_input"
            ]
            == sample_lib.data_model.SampleInputType
        )

    # Test that if a ParseError was raised with docstring.parsers, but no type annotation provided, we cannot deduct type and return None
    def run_fn(some_input: str, some_input_2):
        pass

    with patch("docstring_parser.parse", side_effect=ParseError("mocked error")):
        assert get_argument_types(run_fn)["some_input"] == str
        assert get_argument_types(run_fn)["some_input_2"] == None


def test_optional_type_annotation():
    """Check that we keep the `Optional` wrapping on input types"""

    def _run(sample_input: Optional[int]):
        pass

    assert get_argument_types(_run)["sample_input"] == Optional[int]

    def _run2(sample_input: Optional[str]):
        """
        Args:
            sample_input: str
                optional string input
        """

    assert get_argument_types(_run2)["sample_input"] == Optional[str]


def test_get_argument_type_from_malformed_docstring():
    """This test tests docstring arg type parsing for docstrings in non-conforming styles
    where the actual type name is hidden in the description"""

    def _run(self, foo):
        """

        Args:
            foo: yadda yadda blah sample_lib.data_model.SampleInputType

        Returns:
            None
        """
        pass

    assert get_argument_types(_run)["foo"] == sample_lib.data_model.SampleInputType


def test_get_args_with_no_annotation():
    """Check that we get arguments with no type annotation supplied"""

    def _run(input_1="hello world"):
        pass

    assert get_argument_types(_run)["input_1"] == str

    def _run_with_docstring(input_1):
        """
        Args:
            input_1: str
                Optional str input
        """
        pass

    assert get_argument_types(_run_with_docstring)["input_1"] == Optional[str]

    def _run_with_known_dm_type(sample_input_type):
        pass

    assert (
        get_argument_types(_run_with_known_dm_type)["sample_input_type"]
        == sample_lib.data_model.SampleInputType
    )

    def _run_with_optional_known_dm_type(sample_input_type):
        """
        Args:
            sample_input_type: blah blah
                Optional input
        """
        pass

    assert (
        get_argument_types(_run_with_optional_known_dm_type)["sample_input_type"]
        == Optional[sample_lib.data_model.SampleInputType]
    )

    def _run_with_default_as_list_of_ints(a_list=[1, 2, 3]):
        pass

    assert get_argument_types(_run_with_default_as_list_of_ints)["a_list"] == List[int]

    def _run_with_default_as_list_of_multiple_types(a_list=[True, "hello", 1]):
        pass

    # We parse it as a random type
    assert get_argument_types(_run_with_default_as_list_of_multiple_types)[
        "a_list"
    ] in [List[str], List[bool], List[int]]


def test_get_args_with_defaults():
    """Check that we get arguments with any default value supplied"""

    def _run(
        a, b: bool, c: int = None, d: str = None, e: int = 5, f: float = 0.5, g=None
    ):
        pass

    assert get_args_with_defaults(_run) == {
        "c": None,
        "d": None,
        "e": 5,
        "f": 0.5,
        "g": None,
    }


def test_get_args_with_known_args():
    """Check that we get arguments with a known arg type supplied"""

    def _run(producer_id):
        pass

    assert (
        get_argument_types(_run)["producer_id"]
        == caikit.core.data_model.producer.ProducerId
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A decoder for files that contain multi-part form data of files.
See the w3 spec here: https://www.w3.org/TR/html401/interact/forms.html#h-17.13.4.2

e.g. files of the form
```
Content-Type: multipart/form-data; boundary=my_boundary

--my_boundary
Content-Disposition: form-data; name=""; filename="foo.json"
Content-Type: application/json

[
    {
        "foo": "bar"
    }
]

--my_boundary--
```

NB: This will also handle files which do not start with a Content-Type header.
If the file starts with a different string, we will assume that is the first boundary string.

This is not meant to be an implementation detail of an HTTP handler, this is only meant to be a
general-purpose file decoder for cases where:
- A multipart form-data request has been serialized to a file
- That form data itself contained one or more files (NB: _not_ form fields)
- We wish to get a stream of those inner file contents back

Currently implemented using the `werkzeug` multipart parser, for ease of implementation.

A known drawback is that this reads the full file before streaming back the parsed file contents.
Reading multi-GB files could be slow, as this will cause serialization back to disk of the
files contained within the multipart request.
"""
# Standard
from typing import Iterator
import dataclasses
import os
import typing

# Third Party
from werkzeug import formparser
from werkzeug.datastructures import FileStorage
import werkzeug.http

# First Party
from alog import alog

# Local
from caikit.core.exceptions import error_handler

log = alog.use_channel("MULTIPART_DECODER")
error = error_handler.get(log)


@dataclasses.dataclass
class Part:
    content_type: str
    filename: str
    fp: typing.IO[bytes]


def is_multipart_file(file) -> bool:
    """Returns true if the file appears to contain a multi-part form data request"""
    log.debug3("Determining if %s is a multipart file", file)
    first_line = _get_first_nonempty_line(file)

    # Either: The beginning of the file starts with a boundary string (must start with --)
    if first_line.startswith("--"):
        log.debug3(
            "Assuming file %s is a multipart file because it begins with --", file
        )
        return True
    # Or: it's parseable as a content-type header with a boundary
    header, options = werkzeug.http.parse_options_header(first_line)
    if "multipart" not in header.lower():
        log.debug(
            "No multipart content header detected in [%s: %s], not a multipart file",
            header,
            options,
        )
        return False
    if "boundary" not in options:
        log.debug(
            "No boundary option provided in content type header [%s: %s], not a multipart file",
            header,
            options,
        )
        return False

    # Cool, should be a multipart file
    return True


def stream_multipart_file(file) -> Iterator[Part]:
    """Returns an iterator of Parts, where each Part comes with a content type and an io reader to
    stream the data from.

    NB: This only yields parts which are files, not other form fields.
    """

    boundary = _get_multipart_boundary(file)

    parser = formparser.MultiPartParser()
    with open(file, "rb") as fp:
        _, files = parser.parse(fp, boundary.encode(), None)

        for value in files.values():
            value: FileStorage
            # Get a readable pointer to the file
            fp = value.stream
            # Requires seeking back to the beginning because this was just written
            fp.seek(0)

            log.debug3(
                "Yielding file %s parsed from multipart file %s", value.filename, file
            )
            yield Part(content_type=value.content_type, filename=value.filename, fp=fp)


def _get_multipart_boundary(file) -> str:
    """Returns the multipart boundary string by looking for it in the first line of the file with
    content. Should only be called if is_multipart_file(file) returns True"""
    first_line = _get_first_nonempty_line(file)
    if first_line.startswith("--"):
        return first_line[2:].rstrip(os.linesep)

    _, options = werkzeug.http.parse_options_header(first_line)
    return options["boundary"]


def _get_first_nonempty_line(file) -> str:
    """Return the first line of the file with content.
    Returns empty string if none exists."""
    with open(file, "r", encoding="utf-8") as fp:
        for line in fp:
            stripped_line = line.strip()
            if stripped_line != "":
                return stripped_line
    return ""



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import io
import os
import tempfile

# Third Party
import pytest

# First Party
import aconfig

# Local
from caikit.core import ModuleConfig, ModuleLoader, ModuleSaver
from caikit.core.modules.decorator import SUPPORTED_LOAD_BACKENDS_VAR_NAME
from caikit.core.registries import module_backend_registry, module_registry

# pylint: disable=import-error
from sample_lib.data_model.sample import SampleInputType

# Unit Test Infrastructure
from tests.conftest import temp_config

# NOTE: We do need to import `reset_backend_types` and `reset_module_distribution_registry` for `reset_globals` to work
from tests.core.helpers import *
import caikit.core

## Helpers #####################################################################


@pytest.fixture
def base_module_instance():
    return caikit.core.ModuleBase()


@pytest.fixture
def model_path(fixtures_dir):
    yield os.path.join(fixtures_dir, "dummy_module")


@pytest.fixture
def model_config(model_path):
    yield ModuleConfig.load(model_path)


DUMMY_MODULE_ID = "foo"


def configure_alternate_backend_impl():
    """Function to register a new backend type and register a module implementation
    of existing caikit.core module"""

    @caikit.core.modules.module(id=DUMMY_MODULE_ID, name="dummy base", version="0.0.1")
    class DummyFoo(caikit.core.ModuleBase):
        pass

    # Register backend type
    backend_types.register_backend_type(MockBackend)

    @caikit.core.modules.module(base_module=DummyFoo, backend_type=backend_types.MOCK)
    class DummyBar(caikit.core.ModuleBase):
        def test_fetching_backend(self):
            return [
                backend
                for backend in configured_backends()
                if backend.backend_type == backend_types.MOCK
            ][0]

    return DummyFoo, DummyBar


## ModuleBase ##################################################################


def test_load_evaluation_dataset(fixtures_dir):
    assert isinstance(
        ModuleBase._load_evaluation_dataset(
            os.path.join(fixtures_dir, "dummy_dataset.json")
        ),
        list,
    )


def test_init_available():
    """Make sure that there is an __init__ that takes no args or kwargs"""
    model = caikit.core.ModuleBase()
    assert isinstance(model, caikit.core.ModuleBase)


def test_underscore_load_not_implemented():
    with pytest.raises(NotImplementedError):
        caikit.core.ModuleBase._load(None)


def test_load_delegates_to_underscore_load(good_model_path):
    @caikit.core.modules.module(id="blah", name="dummy base", version="0.0.1")
    class DummyFoo(caikit.core.ModuleBase):
        @classmethod
        def _load(cls, module_loader, **kwargs):
            assert isinstance(module_loader, ModuleLoader)
            return "foo"

    assert DummyFoo.load(good_model_path) == "foo"


def test_run_not_implemented(base_module_instance):
    with pytest.raises(NotImplementedError):
        base_module_instance.run()


def test_save_not_implemented(base_module_instance):
    with pytest.raises(NotImplementedError):
        base_module_instance.save("dummy_path")


def test_train_not_implemented():
    with pytest.raises(NotImplementedError):
        caikit.core.ModuleBase.train()


def test_run_batch_keyword_only(model_path):
    """Make sure that calling run_batch without any positional args is safe"""
    dummy_model = caikit.core.load(model_path)
    dummy_model.run(sample_input=SampleInputType(name="Gabe"))
    dummy_model.run_batch(sample_input=[SampleInputType(name="Gabe")])


## ModuleSaver ################################################################


def test_save_module(model_path):
    dummy_model = caikit.core.load(model_path)

    with tempfile.TemporaryDirectory() as tempdir:
        with ModuleSaver(
            dummy_model,
            model_path=tempdir,
        ) as module_saver:
            module_saver.save_module(dummy_model, "dummy")

            assert os.path.exists(os.path.join(tempdir, "dummy"))
            assert module_saver.config.get("module_paths") is not None
            assert module_saver.config.get("module_paths") == {"dummy": "./dummy"}


def test_save_module_kwargs_get_piped_through(model_path):
    """Testing additional keywords passed to save_module are processed by the module"""
    dummy_model = caikit.core.load(model_path)

    with tempfile.TemporaryDirectory() as tempdir:
        with ModuleSaver(
            dummy_model,
            model_path=tempdir,
        ) as module_saver:
            module_saver.save_module(dummy_model, "dummy", keyword=True)
            saved_model = caikit.core.load(os.path.join(tempdir, "dummy"))
            assert saved_model.metadata["test_keyword"]


def test_save_module_list(model_path):
    """Test valid module list save with a list of models."""
    dummy_model = caikit.core.load(model_path)
    dummy_models_with_rel_path = {"dummy_path": dummy_model}

    with tempfile.TemporaryDirectory() as tempdir:
        with ModuleSaver(
            dummy_model,
            model_path=tempdir,
        ) as module_saver:
            list_of_rel_path, _ = module_saver.save_module_list(
                dummy_models_with_rel_path, "dummy"
            )
            # Ensure that we only get one relative path back
            assert len(list_of_rel_path) == 1
            # And that if we split our path components up, get 2 parts...
            saved_subpath = os.path.split(list_of_rel_path[0])
            assert len(saved_subpath) == 2
            # ...Where the first resolves to the relative path [no hierarchy expected here]
            assert not saved_subpath[0].strip(".")
            # ...And the second is the dummy_path we specified
            assert saved_subpath[1] == "dummy_path"


def test_save_module_list_kwargs_get_piped_through(model_path):
    """Check that additional keywords passed to save_module_list are processed by the module"""
    dummy_model = caikit.core.load(model_path)
    dummy_models_with_rel_path = {"dummy_path": dummy_model}

    with tempfile.TemporaryDirectory() as tempdir:
        with ModuleSaver(
            dummy_model,
            model_path=tempdir,
        ) as module_saver:
            module_saver.save_module_list(
                dummy_models_with_rel_path, "dummy", keyword=True
            )
            saved_model = caikit.core.load(os.path.join(tempdir, "dummy_path"))
            assert saved_model.metadata["test_keyword"]


## ModuleConfig ################################################################


def test_isinstance(model_config):
    assert isinstance(model_config, ModuleConfig)
    assert isinstance(model_config, aconfig.Config)


def test_init_and_members():
    config = ModuleConfig(
        {
            "module_id": "123",
            "module_class": "caikit.core.modules.dummy.Dummy",
            "name": "Dummy Block",
            "string": "hello",
            "integer": 1,
            "float": 0.5,
            "nested": {
                "string": "world",
                "integer": 2,
                "float": -0.123,
            },
        }
    )

    assert config.module_id == "123"
    assert config.module_class == "caikit.core.modules.dummy.Dummy"
    assert isinstance(config.string, str)
    assert config.string == "hello"
    assert isinstance(config.integer, int)
    assert config.integer == 1
    assert isinstance(config.float, float)
    assert config.float == 0.5
    assert isinstance(config.nested, dict)
    assert config.nested.string == "world"
    assert config.nested.integer == 2
    assert config.nested.float == -0.123


def test_backwards_compatibility_for_block():
    config = ModuleConfig({"block_id": "123"})
    assert config.module_id == "123"


def test_backwards_compatibility_for_workflow():
    config = ModuleConfig(
        {
            "workflow_id": "456",
        }
    )
    assert config.module_id == "456"


def test_backwards_compatibility_for_resource():
    config = ModuleConfig(
        {
            "resource_id": "789",
        }
    )
    assert config.module_id == "789"


def test_reserved_keys():
    for reserved_key in ("model_path",):
        with pytest.raises(KeyError):
            ModuleConfig(
                {
                    reserved_key: "x",
                    "module_class": "caikit.core.modules.dummy.Dummy",
                    "name": "Dummy Module",
                }
            )


def test_no_config_yaml():
    with pytest.raises(FileNotFoundError):
        with tempfile.TemporaryDirectory() as tempd:
            ModuleConfig.load(tempd)


def test_model_path_is_file():
    with tempfile.TemporaryDirectory() as tempd:
        tempf = os.path.join(tempd, "junk")

        with open(tempf, mode="w", encoding="utf-8") as fh:
            fh.write("\n")

        with pytest.raises(FileNotFoundError):
            ModuleConfig.load(tempf)


def test_save_and_load(model_config):
    with tempfile.TemporaryDirectory() as tempd:
        model_config.save(tempd)
        ModuleConfig.load(tempd)


## Model Conversion Operations #################################################
# Tests for interacting with file-like objects


def test_model_file_like_conversion(model_path):
    """Test that we can convert a model into a file-like object."""
    dummy_model = caikit.core.load(model_path)
    file_like = dummy_model.as_file_like_object()
    assert isinstance(file_like, io.BytesIO)


def test_model_byte_conversion(model_path):
    """Test that we can convert a model into bytes."""
    dummy_model = caikit.core.load(model_path)
    bytes_like = dummy_model.as_bytes()
    assert isinstance(bytes_like, bytes)


def test_load_file_like_conversion_and_back(model_path):
    """Test that we can load a model, export to a file-like object, and reload the model."""
    dummy_model = caikit.core.load(model_path)
    file_like = dummy_model.as_file_like_object()
    reloaded_model = caikit.core.load(file_like)
    assert isinstance(reloaded_model, caikit.core.ModuleBase)


def test_load_bytes_conversion_and_back(model_path):
    """Test that we can load a model, export to a bytes object, and reload the model."""
    dummy_model = caikit.core.load(model_path)
    bytes_like = dummy_model.as_bytes()
    reloaded_model = caikit.core.load(bytes_like)
    assert isinstance(reloaded_model, caikit.core.ModuleBase)


## Non-Local Backends ##########################################################


def test_can_get_module_id(reset_globals):
    """Test we can get registered module from registry"""
    _, DummyBar = configure_alternate_backend_impl()
    assert (
        module_backend_registry()
        .get(DUMMY_MODULE_ID)
        .get(MockBackend.backend_type)
        .impl_class
        == DummyBar
    )


def test_module_id_registration_for_backend(reset_globals):
    """Test module id is properly registered for backend implementation"""
    DummyFoo, DummyBar = configure_alternate_backend_impl()
    assert DummyBar.MODULE_ID == DummyFoo.MODULE_ID


def test_module_registry_for_backend_type(reset_globals):
    """Test backend is properly registered in the module registry"""
    _ = configure_alternate_backend_impl()
    assert DUMMY_MODULE_ID in module_backend_registry()
    assert MockBackend.backend_type in module_backend_registry()[DUMMY_MODULE_ID]


def test_duplicate_registration_raises(reset_globals):
    """Test duplicate registration of same backend for same module raises"""
    DummyFoo, _ = configure_alternate_backend_impl()
    with pytest.raises(AssertionError):

        @caikit.core.modules.module(
            base_module=DummyFoo, backend_type=backend_types.MOCK
        )
        class DummyBat(caikit.core.ModuleBase):
            pass


def test_backend_impl_inheritance_error(reset_globals):
    """Test creating a backend implementation of a module
    which inherits from base module raises"""
    DummyFoo, DummyBar = configure_alternate_backend_impl()
    with pytest.raises(TypeError):

        @caikit.core.modules.module(backend_type=backend_types.MOCK)
        class DummyBat(DummyFoo):
            pass


def test_class_attributes(reset_globals):
    """Test that the wrapped class has the correct attributes set"""
    DummyFoo, DummyBar = configure_alternate_backend_impl()

    # Make sure all of the right class attrs are set
    assert DummyBar.BACKEND_TYPE == backend_types.MOCK
    assert DummyBar.MODULE_ID == DummyFoo.MODULE_ID
    assert DummyBar.MODULE_NAME
    assert DummyBar.MODULE_VERSION
    assert DummyBar.MODULE_CLASS
    assert DummyBar.PRODUCER_ID

    # Configure and make sure it can be fetched by the class
    with temp_config(
        {
            "model_management": {
                "initializers": {
                    "default": {
                        "type": "LOCAL",
                        "config": {
                            "backend_priority": [
                                {"type": backend_types.MOCK},
                            ]
                        },
                    },
                }
            }
        }
    ):
        assert DummyBar.BACKEND_TYPE == backend_types.MOCK

        # Normally non-local backend modules are loaded through caikit.load, so
        # we need to make sure the local loader has been properly configured to
        # enable local instance instantiation.
        assert MODEL_MANAGER.get_initializer("default")

        # Make sure an instance can fetch via get_backend()
        inst = DummyBar()
        assert inst.test_fetching_backend().backend_type == backend_types.MOCK


def test_default_load_supported_backend(reset_globals):
    """Test the defined backend gets added as the supported backend by default"""
    _, DummyBar = configure_alternate_backend_impl()
    assert hasattr(DummyBar, SUPPORTED_LOAD_BACKENDS_VAR_NAME)
    assert len(DummyBar.SUPPORTED_LOAD_BACKENDS) == 1
    assert DummyBar.SUPPORTED_LOAD_BACKENDS[0] == backend_types.MOCK


def test_override_load_supported_backend(reset_globals):
    """Test if the class can successfully define its own backends
    that it supports load from"""

    @caikit.core.modules.module(id=DUMMY_MODULE_ID, name="dummy base", version="0.0.1")
    class DummyFoo(caikit.core.ModuleBase):
        pass

    class BazBackend(BackendBase):
        backend_type = "BAZ"

    class FooBackend(BackendBase):
        backend_type = "FOO"

    backend_types.register_backend_type(BazBackend)
    backend_types.register_backend_type(FooBackend)

    supported_backends = [backend_types.BAZ, backend_types.FOO]

    @caikit.core.modules.module(base_module=DummyFoo, backend_type=backend_types.BAZ)
    class DummyBaz(caikit.core.ModuleBase):
        SUPPORTED_LOAD_BACKENDS = supported_backends

    assert hasattr(DummyBaz, SUPPORTED_LOAD_BACKENDS_VAR_NAME)
    assert DummyBaz.SUPPORTED_LOAD_BACKENDS == supported_backends


def test_base_module_in_decorator(reset_globals):
    """Test multiple backend implementation of a module doesn't override base module
    from module registry"""

    class BazBackend(BackendBase):
        backend_type = "BAZ"

    class FooBackend(BackendBase):
        backend_type = "FOO"

    backend_types.register_backend_type(BazBackend)
    backend_types.register_backend_type(FooBackend)

    @caikit.core.modules.module(id=DUMMY_MODULE_ID, name="dummy base", version="0.0.1")
    class DummyLocal(caikit.core.ModuleBase):
        pass

    @caikit.core.modules.module(backend_type=backend_types.BAZ, base_module=DummyLocal)
    class DummyBaz(caikit.core.ModuleBase):
        pass

    @caikit.core.modules.module(
        backend_type=backend_types.FOO, base_module=DummyLocal.MODULE_ID
    )
    class DummyFoo(caikit.core.ModuleBase):
        pass

    assert DummyLocal in list(module_registry().values())
